module top (
  input  pin_clk,

  inout  pin_usbp,
  inout  pin_usbn,
  output pin_pu,

  output pin_led

  );

  ////////////////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////
  ////////
  //////// generate 48 mhz clock
  ////////
  ////////////////////////////////////////////////////////////////////////////////
  ////////////////////////////////////////////////////////////////////////////////
  wire clk_48mhz;

  SB_PLL40_CORE #(
    .DIVR(4'b0000),
    .DIVF(7'b0101111),
    .DIVQ(3'b100),
    .FILTER_RANGE(3'b001),
    .FEEDBACK_PATH("SIMPLE"),
    .DELAY_ADJUSTMENT_MODE_FEEDBACK("FIXED"),
    .FDA_FEEDBACK(4'b0000),
    .DELAY_ADJUSTMENT_MODE_RELATIVE("FIXED"),
    .FDA_RELATIVE(4'b0000),
    .SHIFTREG_DIV_MODE(2'b00),
    .PLLOUT_SELECT("GENCLK"),
    .ENABLE_ICEGATE(1'b0)
  ) usb_pll_inst (
    .REFERENCECLK(pin_clk),
    .PLLOUTCORE(clk_48mhz),
    .PLLOUTGLOBAL(),
    .EXTFEEDBACK(),
    .DYNAMICDELAY(),
    .RESETB(1'b1),
    .BYPASS(1'b0),
    .LATCHINPUTVALUE(),
    .LOCK(),
    .SDI(),
    .SDO(),
    .SCLK()
  );

  // Generate reset signal (48Mhz clock domain)
  reg   [13:0] reset_cnt = 0;
  wire  resetn = &reset_cnt;
  always @(posedge clk_48mhz) reset_cnt <= reset_cnt + !resetn;

  // Generate the slow speed clock
  localparam slow_clock_size = 22;
  reg [slow_clock_size:0]  slow_clock;
  wire clk_a1hz;
  always @(posedge clk_48mhz) slow_clock <= slow_clock + 1;
  assign clk_a1hz = slow_clock[slow_clock_size];

  // Generate the slow speed reset signal
  reg   [2:0] slow_reset_cnt = 0;
  wire  slow_resetn = &slow_reset_cnt;
  always @(posedge clk_a1hz) slow_reset_cnt <= slow_reset_cnt + !slow_resetn;

  // Make the LED pin beat on the slow clock
  assign pin_led = usb_p_in;//clk_a1hz & slow_resetn;

  reg   [7:0] uart_di;      // Data from the UART
  wire  [7:0] uart_do;      // Data to the UART
  reg   uart_re, uart_we;   // Control signals to the UART for read and write
  wire  uart_wait;          // Control signal from the UART
  wire  uart_data_ready;    // Control signal from the UART

// States for the debugger

  localparam base_state = 0;
  localparam do_read = 1;
  localparam do_write = 2;
  localparam finish_read_write = 3;

  reg [2:0] state = base_state;

  localparam debug_bytes = 4; // must be power of two
  localparam debug_bytes_l2 = $clog2(debug_bytes);


  reg [7:0] data[0:debug_bytes - 1];
  wire [7:0] data_wd[0:debug_bytes - 1];
  reg [debug_bytes_l2:0] byte_output_count;
  reg [7:0] input_byte;

  // Test
  always @(posedge clk_a1hz) begin
    data[0] <= data[0] + 1; // Delete this if you don't want a slow speed cycle counter, but it's highly recommended.

    data[1] <= data_wd[1]; 
    data[2] <= data_wd[2];
    data[3] <= data_wd[3];
  end

  cpu the_cpu(clk_a1hz, slow_resetn, data_wd[1], data_wd[2], data_wd[3]);


  reg have_read = 0;
  always @(posedge clk_48mhz) begin
    uart_re <= 0;
    uart_we <= 0;
    if (!resetn) begin
        state <= base_state;
        byte_output_count <= 0;
    end
    else begin
        // Always see if there is a ready byte.  If there is we'll latch onto it later.
        if (uart_data_ready) begin
            have_read <= 1;
        end

        // Basic state machine: read a byte if available, otherwise output next debug byte
        case (state)
        base_state: if (!uart_wait) begin
           if (have_read) begin
                uart_re <= 0;
                state <= do_read;
            end else begin
               //data[byte_output_count] <= byte_output_count;
               state <= do_write;
               end
        end
        do_read: begin
            input_byte <= uart_do;
            //data[byte_output_count] <= uart_do;
            uart_re <= 1;
            have_read <= 0;
            state <= finish_read_write;
        end
        do_write: begin
            if (byte_output_count[debug_bytes_l2])
              uart_di <= 8'b11111111;
            else
              uart_di <= data[byte_output_count[debug_bytes_l2 - 1:0]];
            byte_output_count <= byte_output_count + 1;
            uart_we <= 1;
            state <= finish_read_write;
        end
        finish_read_write: begin
            if (!uart_wait)
                state <= base_state;
        end
        endcase 
    end
end

  wire usb_p_tx;
  wire usb_n_tx;
  wire usb_p_rx;
  wire usb_n_rx;
  wire usb_tx_en;
  wire usb_p_in;
  wire usb_n_in;

  // usb uart
  usb_uart uart (
    .clk_48mhz  (clk_48mhz),
    .resetn     (resetn),

    .usb_p_tx(usb_p_tx),
    .usb_n_tx(usb_n_tx),
    .usb_p_rx(usb_p_rx),
    .usb_n_rx(usb_n_rx),
    .usb_tx_en(usb_tx_en),

    .uart_we  (uart_we),
    .uart_re  (uart_re),
    .uart_di  (uart_di),
    .uart_do  (uart_do),
    .uart_wait(uart_wait),
    .uart_data_ready(uart_data_ready)
  );

  assign pin_pu = 1'b1;
  assign usb_p_rx = usb_tx_en ? 1'b1 : usb_p_in;
  assign usb_n_rx = usb_tx_en ? 1'b0 : usb_n_in;

  SB_IO #(
    .PIN_TYPE(6'b 1010_01), // PIN_OUTPUT_TRISTATE - PIN_INPUT
    .PULLUP(1'b 0)
  ) 
  iobuf_usbp (
    .PACKAGE_PIN(pin_usbp),
    .OUTPUT_ENABLE(usb_tx_en),
    .D_OUT_0(usb_p_tx),
    .D_IN_0(usb_p_in)
  );

  SB_IO #(
    .PIN_TYPE(6'b 1010_01), // PIN_OUTPUT_TRISTATE - PIN_INPUT
    .PULLUP(1'b 0)
  ) 
  iobuf_usbn (
    .PACKAGE_PIN(pin_usbn),
    .OUTPUT_ENABLE(usb_tx_en),
    .D_OUT_0(usb_n_tx),
    .D_IN_0(usb_n_in)
  );

endmodule
